/* vi: set tabstop=4 : */

/*
 * Version 1.5 September 1988
 *
 * sp - spell word
 *
 * Usage:	sp [-f dictionary-list] [-acetv] [word ...]
 *
 * Compute the Soundex code for each word on the command line
 * (or each word on the standard input) and compare against a
 * dictionary
 *
 * If the environment variable SP is present, it is assumed to contain command
 * line flags
 *
 * The soundex dictionary list may be specified on the command line
 * The environment variable SPPATH may be set to a list of colon
 * separated pathnames of soundex dictionaries.
 * If a command line dictionary-list (a colon separated list of pathnames) is
 * given in addition to the SPPATH variable, all dictionaries are used.
 *
 * To reduce the size of the word list, certain heuristics are used:
 * the -a option causes all words matched to be printed
 * The output is alphabetically sorted and indicators are printed
 * beside each word:
 *	X   == exact match
 *	!   == close match
 *	*   == near match
 * ' '  == matched
 *
 * Note that the maximum number of colliding words is MAXCOUNT due to the
 * data structure used.
 *
 * Permission is given to copy or distribute this program provided you
 * do not remove this header or make money off of the program.
 *
 * Please send comments and suggestions to:
 *
 * Barry Brachman
 * Dept. of Computer Science
 * Univ. of British Columbia
 * Vancouver, B.C. V6T 1W5
 *
 * .. {alberta, uw-beaver}!ubc-vision!ubc-cs!brachman
 * brachman@cs.ubc.cdn
 * brachman%ubc.csnet@csnet-relay.arpa
 * brachman@ubc.csnet
 */

#include <sys/types.h>
#include <sys/file.h>
#include <sgtty.h>
#include <ctype.h>
#include <stdio.h>

#ifdef NEWDBM
#include <ndbm.h>
#else !NEWDBM
#include <dbm.h>
#endif NEWDBM

#include "sp.h"

#define streq(X, Y)		(!strcmp(X, Y))
#define range(S)		((strlen(S) + 4) / 5)

#define TABLETHRESH		10

datum FETCH();

struct dict_list {
	char *dict;
	struct dict_list *next;
};
struct dict_list *dict_list;
int ndicts;

#define NOMATCH		0
#define BANG		1
#define BLANK		2
#define STAR		3
#define EXACT		4
char matchchar[] = { '?', '!', ' ', '*', 'X' };
typedef short Matchtype;
Matchtype choose();

#define MINALLOC	16		/* initial allocation of wordent's */
struct wordent {
	char *word;				/* word retrieved */
	Matchtype mt;			/* how well it matched user's word */
	short wnum;				/* word number used with tabular printing */
};
struct wordent *wordlist;	/* start of vector of malloc'ed wordents */
int nalloc;					/* number of wordents allocated */
int nmatched;				/* number of wordents used */
int nfound;					/* total number of words found by apply() */

/*
 * Soundex codes
 * The program depends upon the numbers zero through six being used
 * but this can easily be changed
 */
char soundex_code_map[26] = {
/***	 A  B  C  D  E  F  G  H  I  J  K  L  M  N  O  P	***/ 
		 0, 1, 2, 3, 0, 1, 2, 0, 0, 2, 2, 4, 5, 5, 0, 1,

/***	 Q  R  S  T  U  V  W  X  Y  Z			***/
		 2, 6, 2, 3, 0, 1, 0, 2, 0, 2
};

char *version = "sp Version 1.5 27-Sep-88 brachman@ubc.csnet";

#define USAGE		"Usage: sp [-f dictionary-list] [-acetv] [word ...]"

int nlines, ncols;			/* lines and columns available on terminal */

int aflag, cflag, eflag, tflag, vflag;

char *malloc(), *realloc();

main(argc, argv)
int argc;
char **argv;
{
	register int i;
	register char *p;
	char *getenv();

	dict_list = (struct dict_list *) NULL;
	ndicts = 0;
	nalloc = 0;
	nmatched = 0;
	nlines = ncols = 0;
	argc--;
	argv++;
	doargs(&argc, &argv);
	if ((p = getenv("SP")) != (char *) NULL) {
		int mkargc;
		char **mkargv, *space;

		if (breakup(p, &mkargv, &space, &mkargc, " ") < 0) {
			(void) fprintf(stderr, "sp: can't scan SP environment variable\n");
			exit(1);
		}
		doargs(&mkargc, &mkargv);
		if (mkargc != 0) {
			(void) fprintf(stderr, "sp: improper SP environment variable\n");
			exit(1);
		}
		free(space);
	}

	if ((p = getenv("SPPATH")) != (char *) NULL)
		add_dict_list(p);
	if (dict_list == (struct dict_list *) NULL)
		add_dict_list(DEFAULT_SPPATH);

	if (vflag) {
		struct dict_list *dptr;

		(void) printf("%s\n", version);
		(void) printf("Using dictionaries:\n");
		dptr = dict_list;
		for (i = 0; i < ndicts; i++, dptr = dptr->next)
			if (strlen(dptr->dict) > 0)
				(void) printf("\t%s\n", dptr->dict);
		(void) printf("With: ");
		if (aflag)
			(void) printf("+aflag");
		if (cflag)
			(void) printf("+cflag");
		if (eflag)
			(void) printf("+eflag");
		if (tflag)
			(void) printf("+tflag");
		if (vflag)
			(void) printf("+vflag");
		(void) printf("\n");
	}

	if (argc) {
		for (i = 0; i < argc; i++) {
			if (!eflag)
				(void) printf("%s:\n", argv[i]);
			apply(argv[i]);
			if (!eflag)
				(void) printf("\n");
		}
	}
	else {
		int ch, len;
		char *w, word[MAXWORDLEN + 2];

		while (1) {
			(void) printf("Word? ");
			w = word;
			if (fgets(w, sizeof(word), stdin) == (char *) NULL) {
				(void) printf("\n");
				break;
			}
			while (*w == ' ')
				w++;
			if (*w == '\n')
				continue;
			len = strlen(w);
			if (w[len - 1] != '\n') {
				(void) fprintf(stderr, "sp: Word too long: %s", word);
				while ((ch = getchar()) != '\n')	/* flush rest of line */
					putc(ch, stderr);
				putc('\n', stderr);
				continue;
			}
			w[--len] = '\0';
			if (len > MAXWORDLEN) {
				(void) fprintf(stderr, "sp: Word too long: %s\n", word);
				continue;
			}

			apply(w);
			if (!eflag)
				(void) printf("\n");
		}
	}
}

/*
 * Scan the given argc/argv type structure extracting command line flags
 */
doargs(pargc, pargv)
int *pargc;
char ***pargv;
{
	register int argc, fflag, i;
	register char **argv;

	argc = *pargc;
	argv = *pargv;
	while (argc > 0 && argv[0][0] == '-') {
		fflag = 0;		/* to break out of following loop... */
		for (i = 1; argv[0][i] != '\0' && fflag == 0; i++) {
			switch (argv[0][i]) {
			case 'a':
				aflag = 1;
				break;
			case 'c':
				cflag = 1;
				break;
			case 'e':
				eflag = 1;
				break;
			case 'f':
				if (argc == 1) {
					(void) fprintf(stderr, "%s\n", USAGE);
					exit(1);
				}
				add_dict_list(argv[1]);
				argc--;
				argv++;
				fflag = 1;		/* break out of loop */
				break;
			case 't':
				tflag = 1;
				break;
			case 'v':
				vflag = 1;
				break;
			default:
				(void) fprintf(stderr, "%s\n", USAGE);
				exit(1);
			}
		}
		argc--, argv++;
	}
	*pargc = argc;
	*pargv = argv;
}

/*
 * Apply the Soundex search for a word to each dictionary in turn
 * Note that 'DBMINIT' opens both the '.dir' and the '.pag' files
 * and we must close them to avoid running out of file descriptors
 *
 * This routine gets called each time a word is looked up and therefore
 * the dbm files may be repeatedly opened and closed.  Since the vast majority
 * of the time this program is invoked for just a single word it doesn't seem
 * worthwhile to do the right thing by saving file descriptors/DBM pointers.
 * There probably won't be more than two dictionaries in use anyway...
 */
apply(word)
char *word;
{
	register int code, i, nodicts;
	struct dict_list *dptr;

	if (nmatched)
		freewords();
	nfound = nmatched = 0;
	if ((code = soundex(word, 3)) == BAD_WORD)
		return;
	nodicts = 1;
	dptr = dict_list;
	for (i = 0; i < ndicts; i++, dptr = dptr->next) {
		if (strlen(dptr->dict) == 0)
			continue;
		if (DBMINIT(dptr->dict, O_RDONLY) != -1) {
			proc(word, code);
			nodicts = 0;
		}
		DBMCLOSE();
	}
	if (nodicts) {
		(void) fprintf(stderr, "sp: Can't open any dictionaries\n");
		exit(1);
	}
	if (vflag && !eflag && nmatched == 0)
		(void) printf("%s: no match\n", word);
	else
		showwords();
}

/*
 * Look the word up in the current dictionary
 * and save all the matches
 * Note that only three digits are of the Soundex code are stored
 * in a dictionary
 */
proc(uword, code)
char *uword;
int code;
{
	register int c;
	register char *p, *q;
	Matchtype mt;
	datum dbm_key, dbm_content;
	spkey_t *key, keyvec[KEYSIZE];
	char *mk_word();

	key = keyvec;
	dbm_key.dptr = (char *) key;
	dbm_key.dsize = KEYSIZE;
	c = 0;
	while (1) {
		if (c > MAXCOUNT) {
			(void) fprintf(stderr, "sp: entry count overflow\n");
			exit(1);
		}
		mk_key(key, code, c);
		dbm_content = FETCH(dbm_key);

		if (dbm_content.dptr == 0)
			break;
		nfound++;

		if (IS_DELETED(dbm_content)) {
			c++;
			continue;
		}

		p = mk_word(dbm_content.dptr, dbm_content.dsize, code);

		if ((mt = choose(uword, p)) == NOMATCH) {
			c++;
			continue;
		}

		if (nmatched == nalloc) {
			if (nalloc == 0) {
				nalloc = 16;
				wordlist = (struct wordent *)
					malloc(nalloc * sizeof(struct wordent));
			}
			else {
				nalloc <<= 1;
				wordlist = (struct wordent *)
					realloc(wordlist, nalloc * sizeof(struct wordent));
			}
			if (wordlist == (struct wordent *) NULL) {
				(void) fprintf(stderr, "sp: Out of space for wordlist\n");
				exit(1);
			}
		}
		(wordlist + nmatched)->mt = mt;
		if ((q = (char *) malloc(strlen(p) + 1)) == (char *) NULL) {
			(void) fprintf(stderr, "sp: Out of space for words\n");
			exit(1);
		}
		(void) strcpy(q, p);
		(wordlist + nmatched++)->word = q;
		c++;
	}
}

/*
 * Select and print those words which we consider
 * to have matched 'word'
 */
Matchtype
choose(uword, word)
char *uword, *word;
{
	register int c, len, wordlen;
	Matchtype mt;

	if (strmatch(uword, word) == 0)
		mt = EXACT;
	else {
		if (eflag)
			mt = NOMATCH;
		else {
			if (soundex(word, 4) == soundex(uword, 4))
				mt = BANG;
			else {
				c = range(uword);
				wordlen = strlen(uword);
				if (aflag
				 && (wordlen < (len = strlen(word)) - c || len > wordlen + c))
					mt = BLANK;
				else if (!cflag)
					mt = STAR;
				else
					mt = NOMATCH;
			}
		}
	}
	return(mt);
}

showwords()
{
	register int i;
	register struct wordent *w;
	int compar(), printentry(), calclength();

	qsort((char *) wordlist, nmatched, sizeof(struct wordent), compar);
	if (tflag && nmatched > TABLETHRESH) {
		for (w = wordlist, i = 1; i <= nmatched; i++, w++)
			w->wnum = i;
		prtable((char *) wordlist, nmatched, 0, 0, printentry, calclength);
	}
	else {
		w = wordlist;
		for (i = 1; i <= nmatched; i++, w++)
			(void) printf("%c%3d. %s\n", matchchar[w->mt], i, w->word);
	}

	if (vflag)
		(void) printf("(%d total matches)\n", nfound);
}

calclength(base, index)
struct wordent *base;
int index;
{

	return(strlen(base[index].word) + 6);
}

printentry(base, index)
struct wordent *base;
int index;
{
	struct wordent *w;

	w = base + index;
	(void) printf("%c%3d. %s", matchchar[w->mt], w->wnum, w->word);
}

/*
 * Compute an 'n' digit Soundex code for 'word' 
 * See mksp.c
 */
soundex(word, n)
register char *word;
int n;
{
	register int c, digit_part, previous_code, soundex_length;
	register char *p, *w;
	char wcopy[MAXWORDLEN + 2];

	(void) strcpy(wcopy, word);
	p = w = wcopy;
	while (*p != '\0') {
		if (isupper(*p))
			*p = tolower(*p);
		p++;
	}
	if (!isalpha(*w)) {
		(void) fprintf(stderr, "sp: Improper word: %s\n", word);
		return(BAD_WORD);
	}
	digit_part = 0;
	soundex_length = 0;
	previous_code = soundex_code_map[*w - 'a'];
	for (p = w + 1; *p != '\0' && soundex_length < n; p++) {
		if (!isalpha(*p))
			continue;
		c = soundex_code_map[*p - 'a'];
		if (c == 0 || previous_code == c) {
			previous_code = c;
			continue;
		}
		digit_part = digit_part * 7 + c;
		previous_code = c;
		soundex_length++;
	}
	while (soundex_length++ < n)
		digit_part *= 7;
	return((digit_part << 5) + *w - 'a');
}

/*
 * Process a path string (environment variable SPPATH, DEFAULT_SPPATH, or an
 * arg) by separating the pathnames into a list of strings
 *
 * No attempt made to ignore duplicate pathnames
 */
add_dict_list(p)
register char *p;
{
	register int len;
	register char *start;
	struct dict_list *dptr;
	static struct dict_list *dict_end = (struct dict_list *) NULL;

	while (*p != '\0') {
		start = p;
		while (*p != ':' && *p != '\0')
			p++;
		if (start == p && *p == ':') {	/* colon with nothing else */
			p++;
			continue;
		}
		dptr = (struct dict_list *) malloc(sizeof(struct dict_list));
		if (dptr == (struct dict_list *) NULL) {
			(void) fprintf(stderr, "sp: Out of dictionary space\n");
			exit(1);
		}
		len = p - start;
		dptr->dict = (char *) malloc((unsigned) (len + 1));
		if (dptr->dict == (char *) NULL) {
			(void) fprintf(stderr, "sp: Out of dictionary space\n");
			exit(1);
		}
		(void) strncpy(dptr->dict, start, len);
		dptr->dict[len] = '\0';
		dptr->next = (struct dict_list *) NULL;
		if (dict_list == (struct dict_list *) NULL)
			dict_list = dict_end = dptr;
		else {
			dict_end->next = dptr;
			dict_end = dptr;
		}
		ndicts++;
	}
}

freewords()
{
	register int i;
	struct wordent *w;

	w = wordlist;
	for (i = 0; i < nmatched; i++, w++)
		free(w->word);
}

compar(p, q)
struct wordent *p, *q;
{

	return(strmatch(p->word, q->word));
/*	return(strcmp(p->word, q->word)); */	/* use if you prefer case sensitive */
}

#define NCOLS	5

/*
 * Routine to print a table
 * Modified from 'ls.c' mods (BJB/83)
 * Arguments:
 *	base	- address of first entry
 *	num     - number of entries
 *	d_cols  - number of columns to use if > 0, "best" size if == 0
 *	width	- max line width if not zero
 *	prentry - address of the routine to call to print the string
 *	length  - address of the routine to call to determine the length
 *		  of string to be printed 
 *
 * prtable and length are called with the the address of the base and
 * an index
 */
prtable(base, num, d_cols, width, prentry, length)
char *base;
int num, d_cols;
int (*prentry)(), (*length)();
{
        register int c, j;
        register int a, b, cols, loc, maxlen, nrows, z;

        if (num == 0)
                return;
	maxlen = get_maxlen(base, num, length) + 1;
	if (d_cols > 0)
		cols = d_cols;
	else if (width == 0)
		cols = get_columns() / maxlen;
	else
		cols = width / maxlen;
	if (cols == 0)
		cols = NCOLS;
        nrows = (num - 1) / cols + 1;
        for (a = 1; a <= nrows; a++) {
                b = c = z = loc = 0;
                for (j = 0; j < num; j++) {
                        c++;
                        if (c >= a + b)
                                break;
                }
                while (j < num) {
                        (*prentry)(base, j);
			loc += (*length)(base, j);
                        z++;
                        b += nrows;
                        for (j++; j < num; j++) {
                                c++;
                                if (c >= a + b)
                                        break;
                        }
                        if (j < num) {
                                while (loc < z * maxlen) {
					(void) printf(" ");
                                        loc++;
                                }
			}
                }
		(void) printf("\n");
        }
}

static int
get_maxlen(base, num, length)
char *base;
int num;
int (*length)();
{
	register int i, len, max;

	max = (*length)(base, 0);
	for (i = 0; i < num; i++) {
		if ((len = (*length)(base, i)) > max)
			max = len;
	}
	return(max);
}

static int
get_columns()
{
	char *term, tbuf[2024];
	char *getenv();
	int tgetnum();

#ifdef TIOCGWINSZ

#ifdef sun
#define winsize ttysize
#define ws_row	ts_lines
#define ws_col	ts_cols
#undef TIOCGWINSZ
#define TIOCGWINSZ	TIOCGSIZE
#endif sun

	struct winsize win;

    if (ioctl(fileno(stdout), TIOCGWINSZ, (char *) &win) >= 0
			&& win.ws_row > 0) {
    	nlines = win.ws_row;
	    return(ncols = win.ws_col);
	}
#endif
	if ((term = getenv("TERM")) == (char *) NULL)
		return(0);
	if (tgetent(tbuf, term) <= 0)
		return(0);
	nlines = tgetnum("li");
	return(tgetnum("co"));
}
